/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "benchmark.h"
#include "config.h"
#include "compiler.h"
#include "error.h"
#include "cmdline.h"
#include "system.h"
#include "vcluster.h"
#include "memusage.h"
#include "timing.h"
#include "stopwatch.h"
#include "environ.h"
#include "format_units.h"
#include "format_print.h"
#if HAVE_VCLUSTER_PORTALS == 1
#include "vcluster_portals.h"
#endif
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cinttypes>
#include <memory>
#include <cassert>
#include <vector>
#include <math.h>
#include <iostream>
#include <sstream>
#include <numeric>
#include <algorithm>
#include <cstdlib>
#include <limits>

namespace linktest
{

int Benchmark::rank() const{
    return cl->rank();
}

int Benchmark::size() const{
    return cl->size();
}

int Benchmark::numHostRanks() const{
    return cl->numHostRanks();
}

int Benchmark::numLocalRanks() const{
    return cl->numLocalRanks();
}

int Benchmark::hostRank() const{
    return cl->hostRank();
}

int Benchmark::localRank() const{
    return cl->localRank();
}

std::shared_ptr<int[]> Benchmark::hostLocalRanks() const{
    return cl->hostLocalRanks();
}

void Benchmark::barrier() const{
    auto err = cl->barrier();
    if (unlikely(err)) {
        fatal("VirtualCluster::barrier() returned %d", err);
    }
}

bool Benchmark::isStressTest() const{
    return (args->min_walltime > 0.0) || (args->min_iterations > 1);
}

bool Benchmark::randomizeTasks() const{ 
    return args->num_randomize_tasks > 0;
}

/*
 * Check if we can stop the test. Via the command line the user can specify a minimum wall-clock
 * time for the test (excluding overhead) as well as a minimum amount of iteration that should
 * be performed.
 */
bool Benchmark::isLastIteration(const double time, const double tstep, const uint64_t finishedIterations) const {
    bool finished = false;
    if (cl->rank() == 0) {
        finished = time + tstep >= args->min_walltime && finishedIterations >= args->min_iterations;
    }
    cl->bcast(0, &finished, 1);
    return finished;
}

int Benchmark::kernel(const int from, const int to, double* const time_per_msg, const bool doBarrier){
    if (args->do_bidir) {
        return cl->linktest_kbipingpong(from, to, *buf1, *buf2, args, time_per_msg);
    } else if (args->do_unidir) {
        if (args->use_multi_buf) {
            auto M = std::max(args->num_msg, args->num_warmup_msg);
            if(args->num_multi_buf==M){
                return cl->linktest_kUniDirMultiBuf(from, to, *buf_multi, *buf2, args, time_per_msg, doBarrier);
            }else{
                return cl->linktest_kUniDirLimitedMultiBuf(from, to, *buf_multi, *buf2, args, time_per_msg, doBarrier);
            }
        } else {
            return cl->linktest_kUniDir(from, to, *buf1, *buf2, args, time_per_msg, doBarrier);
        }
    } else {
        return cl->linktest_kpingpong(from, to, *buf1, args, time_per_msg);
    }
}

inline unsigned int Benchmark::getFrom(int i, int j, int step) const {
    return std::min(i,j); // Result direction differs from Dorian's implementation
}

inline unsigned int Benchmark::getTo(int i, int j, int step) const {
    return std::max(i,j); // Result direction differs from Dorian's implementation
}

inline unsigned int Benchmark::getNumSteps() const {
    if(args->do_bisection!=0){
        if(args->do_group_processes_by_hostname!=0) return numHostRanks()*numLocalRanks()/2;
        else return size()/2;
    }else if(args->do_group_processes_by_hostname!=0) return (numHostRanks()-1)*numLocalRanks();
    else return size()-1;
}

const int INVALID_RANK = -8; // Random negative Number
const int INVALID_STEP = -9; // Random negative Number

void Benchmark::computeRankStepMappings() {
    stepPermutation.resize(getNumSteps());
    getRankFromStep.resize(getNumSteps());
    getStepFromRank.resize(size());

    std::unique_ptr<StopwatchI> rootWatch = Stopwatchfactory::getRootWatch(cl->rank());

    rootWatch->start();
    if(args->do_mix) {
        if( rank() == 0 ) {
            std::iota(stepPermutation.begin(), stepPermutation.end(), 0);
            std::shuffle(stepPermutation.begin(), stepPermutation.end(), randomNumberEngineSteps);
        }
        cl->bcast(0, stepPermutation.data(), getNumSteps());
    }else{
        std::iota(stepPermutation.begin(), stepPermutation.end(), 0);
    }
    rootWatch->stop();
    printTimingIfRoot(cl->rank(), "[randvec]", rootWatch->getDuration());

    rootWatch->start();
    std::fill(getRankFromStep.begin(), getRankFromStep.end(), INVALID_RANK);
    std::fill(getStepFromRank.begin(), getStepFromRank.end(), INVALID_STEP);
    if(args->do_bisection) {
        mapWithBisection(args->num_randomize_tasks > 0, args->do_group_processes_by_hostname != 0);
    } else {
        mapWith1FactorAlgorithm(args->num_randomize_tasks > 0, args->do_group_processes_by_hostname != 0);
    }
    rootWatch->stop();
    printTimingIfRoot(cl->rank(), "[getpart]", rootWatch->getDuration());

#ifdef DEBUG_PRINT_COMMUNICATION_SCHEME
    if( rank() == 0) {
        std::stringstream buf;
        buf << "getRankFromStep: " << getRankFromStep;
        debug("[%3d] %s", 0, buf.str().c_str());buf.str(std::string());
        buf << "getStepFromRank: " << getStepFromRank;
        debug("[%3d] %s", 0, buf.str().c_str());buf.str(std::string());
        buf << "stepPermutation: " << stepPermutation;
        debug("[%3d] %s", 0, buf.str().c_str());buf.str(std::string());
        std::vector<int> buf1;buf1.resize(getNumSteps());
        std::vector<int> buf2;buf2.resize(size()       );
        for(int i = 1; i < size(); i++) {
            cl->recv(i, buf1.data(), buf1.size());
            buf << "getRankFromStep: " << buf1;
            debug("[%3d] %s", i, buf.str().c_str());buf.str(std::string());
            cl->recv(i, buf2.data(), buf2.size());
            buf << "getStepFromRank: " << buf2;
            debug("[%3d] %s", i, buf.str().c_str());buf.str(std::string());
            cl->recv(i, buf1.data(), buf1.size());
            buf << "stepPermutation: " << buf1;
            debug("[%3d] %s", i, buf.str().c_str());buf.str(std::string());
        }
    } else {
        cl->send(0, getRankFromStep.data(), getRankFromStep.size());
        cl->send(0, getStepFromRank.data(), getStepFromRank.size());
        cl->send(0, stepPermutation.data(), stepPermutation.size());
    }
    barrier();
#endif
}

std::vector<int> Benchmark::oneFactorAlgorithm(const int size, const int rank) {
    // We use the all-to-all 1-factor algorithm taken from:
    // https://en.wikipedia.org/wiki/All-to-all_(parallel_pattern)#1-factor_algorithm
    // Date: 11.07.2022 (last checked)

    const int p = size;
    const int j = rank;

    std::vector<int> rankFromStep(p-1); // mapping from step to communication partner rank

    for(int i = 0; i < p-1; i++) {
        int idle = (p*i/2) % (p-1); // PE that would be idle in p is uneven case
        if( j == p-1 ) {
            rankFromStep.at(i) = idle;
        } else {
            if ( j == idle ) {
                rankFromStep.at(i) = p-1;
            } else {
                rankFromStep.at(i) = (i-j+p-1) % (p-1);
            }
        }
    }

    return rankFromStep;
}

void Benchmark::mapWith1FactorAlgorithm(bool randomizeTasks, bool group_processes_by_hostname) {
    // We use the 1-factor algorithm for parallel communications to reduces congestion
    // Implementation (including names) taken from
    // https://en.wikipedia.org/wiki/All-to-all_(parallel_pattern)#1-factor_algorithm

    const int numRanks=(group_processes_by_hostname?numHostRanks():size());

    std::vector<int> randomrankFromRank(numRanks);
    std::vector<int> randomLocalRankFromLocalRank;
    if(group_processes_by_hostname) randomLocalRankFromLocalRank.resize(numLocalRanks());
    if(randomizeTasks) {
        if(rank() == 0){
            std::iota(randomrankFromRank.begin(), randomrankFromRank.end(), 0);
            std::shuffle(randomrankFromRank.begin(), randomrankFromRank.end(), randomNumberEngineTasks);
            if(group_processes_by_hostname){
                std::iota(randomLocalRankFromLocalRank.begin(), randomLocalRankFromLocalRank.end(), 0);
                std::shuffle(randomLocalRankFromLocalRank.begin(), randomLocalRankFromLocalRank.end(), randomNumberEngineTasks);
            }
        }
        cl->bcast(0, randomrankFromRank.data(), numRanks);
        if(group_processes_by_hostname) cl->bcast(0, randomLocalRankFromLocalRank.data(), numLocalRanks());
    }else{
        std::iota(randomrankFromRank.begin(), randomrankFromRank.end(), 0);
        if(group_processes_by_hostname) std::iota(randomLocalRankFromLocalRank.begin(), randomLocalRankFromLocalRank.end(), 0);;
    }
    std::vector<int> rankFromRandomrank(numRanks);
    for(std::size_t i = 0; i < randomrankFromRank.size(); i++) {
        rankFromRandomrank[randomrankFromRank[i]] = i;
    }

    if(group_processes_by_hostname){
        auto randomRankFromStep = oneFactorAlgorithm(numRanks, randomrankFromRank[hostRank()]);
        int step=0;
        for (int hRank = 0; hRank < numRanks-1; hRank++) { //Iterate through host ranks
            const int partnerHostRank = rankFromRandomrank[randomRankFromStep[hRank]];
            for (int lRank = 0; lRank < numLocalRanks(); lRank++) { //Iterate through local ranks
                const int partnerRank = hostLocalRanks().get()[partnerHostRank*numLocalRanks()+(numLocalRanks()+(localRank()+(partnerHostRank>hostRank()?1:-1)*randomLocalRankFromLocalRank[lRank]))%numLocalRanks()];
                getRankFromStep.at(step) = partnerRank;
                getStepFromRank.at(partnerRank) = step;
                step++;
            }
        }
    }else{
        auto randomRankFromStep= oneFactorAlgorithm(size(), randomrankFromRank[rank()]);
        for (int step = 0; step < numRanks - 1; step++) {
            const int partnerRank = rankFromRandomrank[randomRankFromStep[step]];
            getRankFromStep.at(step) = partnerRank;
            getStepFromRank.at(partnerRank) = step;
        }
    }
}

void Benchmark::mapWithBisection(bool randomizeTasks, bool group_processes_by_hostname) {
    const int n = (group_processes_by_hostname ? numHostRanks()/2 : size()/2);
    const int r = (group_processes_by_hostname ? hostRank() : rank());
    std::vector<int> randomRankFromRank(n);
    std::vector<int> randomLocalRankFromLocalRank;
    if(group_processes_by_hostname) randomLocalRankFromLocalRank.resize(numLocalRanks());
    if(randomizeTasks){
        if( rank() == 0 ){
            std::iota(randomRankFromRank.begin(), randomRankFromRank.end(), 0);
            std::shuffle(randomRankFromRank.begin(), randomRankFromRank.end(), randomNumberEngineTasks);
            if(group_processes_by_hostname){
                std::iota(randomLocalRankFromLocalRank.begin(), randomLocalRankFromLocalRank.end(), 0);
                std::shuffle(randomLocalRankFromLocalRank.begin(), randomLocalRankFromLocalRank.end(), randomNumberEngineTasks);
            }
        }
        cl->bcast(0, randomRankFromRank.data(), n);
        if(group_processes_by_hostname) cl->bcast(0, randomLocalRankFromLocalRank.data(), numLocalRanks());
    }else{
        std::iota(randomRankFromRank.begin(), randomRankFromRank.end(), 0);
        if(group_processes_by_hostname) std::iota(randomLocalRankFromLocalRank.begin(), randomLocalRankFromLocalRank.end(), 0);
    }
    if(group_processes_by_hostname){
        int k=0;
        for(int hRank=0; hRank<n; hRank++ ){
            const int partnerHostRank=(r<n?n+(n-1-r+randomRankFromRank[hRank])%n:(2*n -1-r+randomRankFromRank[hRank])%n);
            for(int lRank=0; lRank<numLocalRanks(); lRank++ ){
                getRankFromStep.at(k)=hostLocalRanks().get()[partnerHostRank*numLocalRanks()+(numLocalRanks()+(localRank()+(partnerHostRank>hostRank()?1:-1)*randomLocalRankFromLocalRank[lRank]))%numLocalRanks()];
                getStepFromRank.at(getRankFromStep.at(k)) = k;
                k++;
            }
        }
    }else{
        for( auto i=0; i<n; i++ ){
            if( r < n ){
                getRankFromStep.at(i) = n + (  n - 1 - r + randomRankFromRank[i]) % n;
            } else {
                getRankFromStep.at(i) =     (2*n - 1 - r + randomRankFromRank[i]) % n;
            }
            getStepFromRank.at(getRankFromStep.at(i)) = i;
        }
    }
}

int Benchmark::printStepResults(const size_t       numStepsProcessed,
                                const size_t       numRemainingIterations,
                                const double       tWork,
                                const double       minTimeForStep,
                                const double       avgTimeForStep,
                                const double       maxTimeForStep,
                                const double       sumBandwidth,
                                const double       runningSumBandwidth,
                                const unsigned int strOffset
                               ){
    const int fieldWidth = 10;
    const int remainingSteps = getNumSteps() - numStepsProcessed;
    const double tavg = tWork / numStepsProcessed;

    std::printf("min: %ss (%sB/s) avg: %ss (%sB/s) max: %ss (%sB/s)",
                UnitPrefix::SI_prefix(minTimeForStep, fieldWidth).get(), UnitPrefix::IEC_prefix(args->len_msg/minTimeForStep, fieldWidth).get(),
                UnitPrefix::SI_prefix(avgTimeForStep, fieldWidth).get(), UnitPrefix::IEC_prefix(args->len_msg/avgTimeForStep, fieldWidth).get(),
                UnitPrefix::SI_prefix(maxTimeForStep, fieldWidth).get(), UnitPrefix::IEC_prefix(args->len_msg/maxTimeForStep, fieldWidth).get()
               );
    if(args->do_bisection) std::printf(" avg. bisection bw: %sB/s\n",UnitPrefix::IEC_prefix(sumBandwidth/2, 10).get());
    else std::printf("\n");
    if(numStepsProcessed>1){
        std::printf("%*s: min: %ss (%sB/s) avg: %ss (%sB/s) max: %ss (%sB/s)",strOffset,"Aggregated Results",
                    UnitPrefix::SI_prefix(stats->min_time_all                  ,fieldWidth).get(), UnitPrefix::IEC_prefix(args->len_msg/stats->min_time_all                  ,fieldWidth).get(),
                    UnitPrefix::SI_prefix(stats->avg_time_all/numStepsProcessed,fieldWidth).get(), UnitPrefix::IEC_prefix(args->len_msg/stats->avg_time_all*numStepsProcessed,fieldWidth).get(),
                    UnitPrefix::SI_prefix(stats->max_time_all                  ,fieldWidth).get(), UnitPrefix::IEC_prefix(args->len_msg/stats->max_time_all                  ,fieldWidth).get()
                   );
        if(args->do_bisection) std::printf(" avg. bisection bw: %sB/s\n",UnitPrefix::IEC_prefix(runningSumBandwidth/(2*numStepsProcessed), 10).get());
        else std::printf("\n");
    }
    std::printf("%*s: Elapsed time: %ss (%ss/step). Estimated time remaining for iteration: %ss (%u step%c)",
                strOffset,"Timing Summary",
                UnitPrefix::SI_prefix(tWork,               10).get(),
                UnitPrefix::SI_prefix(tavg,                10).get(),
                UnitPrefix::SI_prefix(remainingSteps*tavg, 10).get(),
                remainingSteps,
                remainingSteps==1 ? 's' : ' '
               );
    if(numRemainingIterations!=std::numeric_limits<std::size_t>::max()){
        const double tRemaining=(1+numRemainingIterations)*remainingSteps*tavg+numRemainingIterations*tWork;
        std::printf(". Estimated total time remaining: %ss\n", UnitPrefix::SI_prefix(tRemaining,fieldWidth).get());
    }else std::printf("\n");
    std::printf("\n");
    std::fflush(stdout);

    return SUCCESS;
}

int Benchmark::printIterationResults(const int iter){
    if(0 == cl->rank()) {
        std::printf("\n\n");
        std::printf("LinkTest Timing Results for Iteration %4d:\n",iter);
        std::printf("RESULT: Min Time: %ss (%sB/s)\n",
                    UnitPrefix::SI_prefix(stats->min_time_all, 13).get(),
                    UnitPrefix::IEC_prefix(args->len_msg/stats->min_time_all, 9).get()
                   );
        std::printf("RESULT: Max Time: %ss (%sB/s)\n",
                    UnitPrefix::SI_prefix(stats->max_time_all, 13).get(),
                    UnitPrefix::IEC_prefix(args->len_msg/stats->max_time_all, 9).get()
                   );
        std::printf("RESULT: Avg Time: %ss (%sB/s)\n",
                    UnitPrefix::SI_prefix(stats->avg_time_all, 13).get(),
                    UnitPrefix::IEC_prefix(args->len_msg/stats->avg_time_all, 9).get()
                   );
        if (args->do_alltoall){
        std::printf("LinkTest All-to-All Timing Results:\n");
        std::printf("RESULT: Min Time: %ss (%sB/s)\n",
                    UnitPrefix::SI_prefix(stats->a2a_min_time_all, 13).get(),
                    UnitPrefix::IEC_prefix(args->len_msg/stats->a2a_min_time_all, 9).get()
                   );
        std::printf("RESULT: Max Time: %ss (%sB/s)\n",
                    UnitPrefix::SI_prefix(stats->a2a_max_time_all, 13).get(),
                    UnitPrefix::IEC_prefix(args->len_msg/stats->a2a_max_time_all, 9).get()
                   );
        std::printf("RESULT: Avg Time: %ss (%sB/s)\n",
                    UnitPrefix::SI_prefix(stats->a2a_avg_time_all, 13).get(),
                    UnitPrefix::IEC_prefix(args->len_msg/stats->a2a_avg_time_all, 9).get()
                   );
        std::printf("RESULT: All-to-all is %8.4f times as fast as the equivalent cumulative point-to-point communication.\n",((cl->size()*(cl->size()-1))*stats->avg_time_all)/stats->a2a_avg_time_all);
        }
    }

    EXEC_NOFAIL(cl->barrier());

    return SUCCESS;
}

/*
 * Runs the all-tol-all kernel if requested.
 */
int Benchmark::work_alltoall(){
    double time, min_time, max_time, avg_time;

    /* Run the all-to-all kernel */
    EXEC_NOFAIL(cl->linktest_kalltoall(*buf_all2all, args, &time));

    /* Determine minimum, maximum & average */
    EXEC_NOFAIL(cl->reduce(0,&min_time,&time,1,MIN_DOUBLE));
    EXEC_NOFAIL(cl->reduce(0,&max_time,&time,1,MAX_DOUBLE));
    EXEC_NOFAIL(cl->reduce(0,&avg_time,&time,1,SUM_DOUBLE));
    avg_time/=cl->size();

    /* Store Data */
    stats->a2a_time+=time;
    if (0 == rank()){
        stats->a2a_min_time_all = std::min(stats->a2a_min_time_all,min_time);
        stats->a2a_max_time_all = std::max(stats->a2a_max_time_all,max_time);
        stats->a2a_avg_time_all+=avg_time;
    }

    return SUCCESS;
}

/*
 * Run the ping-pong kernel in both directions.
 */
int Benchmark::work_pingpong_parallel(const int partner,const int sign, double* const time_per_msg){
    double tmp1, tmp2;

    auto from = (sign < 0) ? partner : rank();
    auto to   = (sign < 0) ? rank() : partner;

    barrier();
    EXEC_NOFAIL(kernel(from, to, &tmp1, true));
    barrier();
    EXEC_NOFAIL(kernel(to, from, &tmp2, true));
    barrier();

    *time_per_msg = (sign > 0) ? tmp1 : tmp2;

    return SUCCESS;
}

/*
 * Run the ping-pong kernel in a serial fashion.
 */
int Benchmark::work_pingpong_serial(const int partner, double* const time_per_msg){
    for (auto i = 0; i < size(); ++i) {
        barrier();
        if (i == rank()) {
            EXEC_NOFAIL(kernel(rank(), partner, time_per_msg, false));
            std::printf("%6d->%6d: %ss (%sB/s)\n",
                i, 
                partner,
                UnitPrefix::SI_prefix(*time_per_msg, 12).get(), 
                UnitPrefix::IEC_prefix(args->len_msg / *time_per_msg, 10).get()
            );
        } else if (i == partner) {
            EXEC_NOFAIL(kernel(partner, rank(), nullptr, false));
        } else {
            #if HAVE_VCLUSTER_PORTALS == 1
            if(cl->nameRef() == VirtualClusterPortals::NAME) {
                // 2 per kernel, warmup kernel and measuring kernel
                barrier();
                barrier();
                barrier();
                barrier();
            }
            #endif
        }
    }
    barrier();

    return SUCCESS;
}

/*
 * Run the ping-pong kernel. All processing elements will be engaged at the
 * same time.
 */
int Benchmark::work_pingpong(const int step,double* const min_time,double* const avg_time,double* const max_time,double* const sum_bandwidth){
    int partner, sign;
    double time_per_msg, sum_time, msg_per_time, sum_msg_per_time;

    time_per_msg = -1.0; // For catching bugs

    partner = getRankFromStep.at(step);
    sign = rank() < partner ? 1 : -1;
    if (likely(!args->do_serial)) {
        EXEC_IFFAIL(work_pingpong_parallel(partner, sign, &time_per_msg), fatal("work_pingpong_parallel failed."));
    } else {
        EXEC_IFFAIL(work_pingpong_serial(partner, &time_per_msg), fatal("work_pingpong_serial failed."));
    }
    stats->accesspattern[partner] = step + 1; // In SION file steps are numbered starting with 1.
    stats->ptimings[partner] = time_per_msg;

    EXEC_NOFAIL(cl->reduce(0,  min_time, &time_per_msg, 1, MIN_DOUBLE));
    EXEC_NOFAIL(cl->reduce(0,  max_time, &time_per_msg, 1, MAX_DOUBLE));
    EXEC_NOFAIL(cl->reduce(0, &sum_time, &time_per_msg, 1, SUM_DOUBLE));

    msg_per_time = 1 / time_per_msg;
    EXEC_NOFAIL(cl->reduce(0, &sum_msg_per_time, &msg_per_time, 1, SUM_DOUBLE));

    if (0 == rank()) {
        *avg_time = sum_time/size();
        *sum_bandwidth = args->len_msg * sum_msg_per_time;

        stats->min_time_all = std::min(stats->min_time_all, *min_time);
        stats->max_time_all = std::max(stats->max_time_all, *max_time);
        stats->avg_time_all += *avg_time;
    }

    return SUCCESS;
}

void Benchmark::finalize_results() {
    stats->avg_time_all /= getNumSteps();
    if (args->do_alltoall) stats->a2a_avg_time_all /= 2;
}

/*
 * Run one iteration of the benchmark
 */
int Benchmark::run_iteration(const std::size_t iter, const std::size_t numRemainingIterationsIterations, double* const tWork, double* const tstep){
    double minTimeForStep, avgTimeForStep, maxTimeForStep, sumBandwidth;
    double runningSumBandwidth=0.0;

    /* Execute all-to-all test if desired */
    barrier();
    if (args->do_alltoall){
        EXEC_NOFAIL(work_alltoall());
    }

    /* Execute point-to-point tests */
    if(rank()==0){
        static const char* fmt[] = {
            "Parallel Results for step %*d: ",
            "Serial   Results for step %*d: "
        };
        const std::size_t nDigits   = number_of_digits_including_minus(getNumSteps());
        const std::size_t strOffset = std::strlen(fmt[!!args->do_serial]) - 5 + nDigits; //-3 removed %*d
        if(strOffset>std::numeric_limits<unsigned int>::max())throw std::overflow_error("data is larger than INT_MAX");
        const unsigned int strLength=static_cast<unsigned int>(strOffset);

        std::size_t j=0;
        for (unsigned int step: stepPermutation) {
            if(!args->do_serial) std::printf(fmt[!!args->do_serial], nDigits, step+1); //Start print out early so user knows step has started
            std::fflush(stdout);
            const double stepStartTime = walltime();
            barrier();
            EXEC_NOFAIL(work_pingpong(step, &minTimeForStep, &avgTimeForStep, &maxTimeForStep, &sumBandwidth));
            *tWork += (walltime() - stepStartTime);
            runningSumBandwidth+=sumBandwidth;
            if(args->do_serial) std::printf(fmt[!!args->do_serial], nDigits, step+1);
            printStepResults(++j, numRemainingIterationsIterations, *tWork, minTimeForStep, avgTimeForStep, maxTimeForStep, sumBandwidth, runningSumBandwidth, strLength);
        }
    }else{
        for (unsigned int step: stepPermutation) {
            barrier();
            EXEC_NOFAIL(work_pingpong(step, &minTimeForStep, &avgTimeForStep, &maxTimeForStep, &sumBandwidth));
        }
    }

    /* Execute all-to-all test if desired */
    barrier();
    if (args->do_alltoall){
        EXEC_NOFAIL(work_alltoall());
    }

    finalize_results();
    return printIterationResults(iter);
}

/*
 * Gather the list of n slowest communication pairs.
 */
int Benchmark::gather_slow_pairs(struct slow_pair* const sp,const int n){
    const std::size_t num_slowest=(std::size_t)n;

    //Construct Vector Wrapper
    //TODO: This vectors is made here, but duplicated in function call.
    //      In general we should maybe either already have this vector
    //      or pass it by reference.
    std::vector<int>    from;from.reserve(cl->size()-1);
    std::vector<int>      to;  to.reserve(cl->size()-1);
    std::vector<double> time;time.reserve(cl->size()-1);
    std::vector<int>    slow_from;
    std::vector<int>    slow_to;
    std::vector<double> slow_time;
    for(std::size_t i=0;i<(std::size_t)cl->size();i++){
        if(std::isnan(stats->ptimings[i]))continue; //NAN timings were not measured
        from.push_back(cl->rank()       );
          to.push_back(i                );
        time.push_back(stats->ptimings[i]);
    }
    find_and_sort_slowest_connections(cl,num_slowest,from,to,time,slow_from,slow_to,slow_time);

    /* Construct Slow Pairs Structure */
    for(std::size_t i=0;i<num_slowest;i++){
        sp[i].from=slow_from[i];
        sp[i].to  =slow_to[i];
        sp[i].time=slow_time[i];
    }

    return SUCCESS;
}

int Benchmark::retest_one_slow_pair(const int from,const int to, double* const time){
    debug("Benchmark::retest_one_slow_pair(%d, %d)",from, to);
    double tv;
    barrier();

    if ((from == rank()) || (to == rank())) {
        EXEC_NOFAIL(kernel(from, to, &tv, false));
    } else { //TODO remove this hack. Seperating MemoryBuffer preparation from kernels. Bann all barriers from kernels
        #if HAVE_VCLUSTER_PORTALS == 1
        if(cl->nameRef() == VirtualClusterPortals::NAME) {
            // 2 per kernel, warmup kernel and measuring kernel
            barrier();
            barrier();
            barrier();
            barrier();
        }
        #endif
    }
    barrier();

    if (0 != from) {
        if (from == rank()) {
            EXEC_NOFAIL(cl->send(0, &tv, 1));
        }
        if (0 == rank()) {
            EXEC_NOFAIL(cl->recv(from, &tv, 1));
        }
    }

    if (0 == rank())
        *time = tv;

    return SUCCESS;
}

int Benchmark::retest_slow_pairs(slow_pair* const sp,const int n,const int iter){
    for (int i = 0; i < n; ++i) {
        double time;

        const int from = sp[i].from;
        const int to   = sp[i].to;

        if (0 == rank()) {
            std::printf(" %6d: Retest %6d  <->  %6d:", i, from, to);
            std::fflush(stdout);
        }

        EXEC_NOFAIL(retest_one_slow_pair(from, to, &time));

        if (0 == rank()) {
            stats->stimings_new[i] = time;
            stats->stimings_old[i] = sp[i].time;
            stats->sfrom[i] = from;
            stats->sto[i] = to;

            std::printf(" 1st: %ss (%sB/s)    2nd: %ss (%sB/s)\n",
                   UnitPrefix::SI_prefix(sp[i].time, 10).get(),UnitPrefix::IEC_prefix(args->len_msg/sp[i].time, 10).get(),
                   UnitPrefix::SI_prefix(      time, 10).get(),UnitPrefix::IEC_prefix(args->len_msg/      time, 10).get()
                  );
            std::fflush(stdout);
        }
    }

    if (0 == rank()) {
        double tmin, tmax, tavg;

        tmin = tmax = tavg = stats->stimings_new[0];

        for (auto i = 1; i < n; ++i) {
            tmin  = std::min(tmin, stats->stimings_new[i]);
            tmax  = std::max(tmax, stats->stimings_new[i]);
            tavg +=           stats->stimings_new[i];
        }
        tavg /= n;

        std::printf("LinkTest Slow-Pairs Results - Iteration %3d:\n",iter);
        std::printf("RESULT: Min Time: %ss (%sB/s)\n",
               UnitPrefix::SI_prefix(tmin, 13).get(),UnitPrefix::IEC_prefix(args->len_msg/tmin, 9).get()
              );
        std::printf("RESULT: Max Time: %ss (%sB/s)\n",
               UnitPrefix::SI_prefix(tmax, 13).get(),UnitPrefix::IEC_prefix(args->len_msg/tmax, 9).get()
              );
        std::printf("RESULT: Avg Time: %ss (%sB/s)\n",
               UnitPrefix::SI_prefix(tavg, 13).get(),UnitPrefix::IEC_prefix(args->len_msg/tavg, 9).get()
              );
        std::fflush(stdout);
    }

    return SUCCESS;
}

int Benchmark::handle_slow_pairs(const int iter){
    const int n = args->max_stest;

    if(n){ //It makes no sense to run the serial tests suit if no tests are to be run
        std::unique_ptr<StopwatchI> rootWatch = Stopwatchfactory::getRootWatch(rank());

        rootWatch->start();
        auto sp = new slow_pair[n];
        auto ret = gather_slow_pairs(sp, n);
        if (unlikely(ret))
            goto out;

        rootWatch->stop();
        printTimingIfRoot(rank(), "[search slow]", rootWatch->getDuration());

        rootWatch->start();
        ret = retest_slow_pairs(sp, n, iter);
        if (unlikely(ret)) goto out;
        rootWatch->stop();
        printTimingIfRoot(rank(), "[test slow]", rootWatch->getDuration());

        delete[] sp;
        return SUCCESS;
        out:
        delete[] sp;
        return ERROR;
    }
    return SUCCESS;
}

int Benchmark::write_to_file(){
    if (args->do_nosion) {
        return ERROR; // currently we can only use SION for writing the output
    } else if (stats != nullptr) {
        if (args->do_sion_par) {
            return cl->write_parallel(args, newStatsVec);
        } else {
            return cl->write_funnelled(args, newStatsVec);
        }
    } else {
        fatal("LinktestStats are not set. Writing out failed");
        return EXIT_FAILURE;
    }
}

/* Parse the command line arguments. A return value of NULL causes main() to
 * exit. This can be caused by invalid command line arguments or by the special
 * arguments like -h (help).
 *
 * In order to be highly portable we assume that command line arguments can
 * only be parsed by the root processing element.
 */
int Benchmark::main_cmdline(){
    /* Fix Command-Line Arguments */
    if(0==cl->rank()){
        args = fixup_cmdline_args((uint64_t)cl->size());
        if(unlikely(!args)){
            error("Failed to correct command-line arguments.");
            return(1);
        }
    }

    /* Broadcast Command-Line Arguments */
    args = bcast_cmdline_args(cl.get(),0,args);

    return(0);
}

int Benchmark::init() {
    // Check if Cluster has even number of ranks
    if(size()%2!=0){
        if(0==cl->rank()){error("An even number of tasks is required for this benchmark.");}
        std::abort();
        return ERROR;
    }

    // Set Allocator
    switch(args->alloc_typ){
        case(AllocatorInvalid):
        case(AllocatorDefault):
            fatal("Encountered invalid memory-buffer allocator!");
            std::abort();
            return ERROR;
            break;
        case(AllocatorAlignedMalloc):
            if(cl->rank()==0){info("Using memory-aligned--malloc allocator");std::fflush(stdout);}
            alloc.reset(new MallocAllocator());
            break;
        case(AllocatorPinnedMmap):
            if(cl->rank()==0){info("Using pinned--memory-map allocator");std::fflush(stdout);}
            alloc.reset(new PinnedMmapAllocator());
            break;
        case(AllocatorPOSIXAlignedMalloc):
            if(cl->rank()==0){info("Using POSIX memory-aligned--malloc allocator");std::fflush(stdout);}
            alloc.reset(new PosixMemAlignedAllocator());
            break;
        case(AllocatorCUDA):
            #if HAVE_VCLUSTER_CUDA == 1
                if(cl->rank()==0){info("Using CUDA memory allocator"); std::fflush(stdout);}
                gpudev.reset(new cuda::GpuDevice(System::singleton()->closest_gpu_device()));
                gpuctx.reset(new cuda::GpuContext(gpudev.get()));
                alloc.reset(new CudaDeviceAllocator(gpuctx.get()));
            #else
                fatal("Requested CUDA memory allocator, but compiled without CUDA support.");
                std::abort();
                return ERROR;
            #endif
            break;
        default:
            fatal("Encountered unknown memory-buffer allocator!");
            std::abort();
            return ERROR;
    }

    if(args->num_randomize_tasks > 0) {
        randomNumberEngineTasks = std::mt19937(args->seed_randomize_tasks);
    } else {
        randomNumberEngineTasks = std::mt19937(0);
    }

    if(args->num_randomize_tasks > 0) {
        randomNumberEngineSteps = std::mt19937(args->seed_randomize_steps);
    } else {
        randomNumberEngineSteps = std::mt19937(0);
    }

    if(args->do_group_processes_by_hostname > 0) {
        cl->getHostAndLocalRank();
    }

    return SUCCESS;
}

void Benchmark::prepareBuffers() {
    auto N = args->len_msg + cl->kpingpong_minimal_buffer_overhead();
    const auto ALIGNMENT = 4096;
    if (args->do_bidir){
        if(args->randomize_buffers){
            buf1.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT,(uint_fast64_t)args->buf_mt_seed));
            buf2.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT,(uint_fast64_t)args->buf_mt_seed));
        }else{
            buf1.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT));
            buf2.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT));
        }
        buf1->fill();
        buf2->fill();
    }else if (args->do_unidir) {
        if (args->use_multi_buf){
            if(args->randomize_buffers){
                buf_multi.reset(new MemoryBufferMulti(alloc.get(),args->num_multi_buf,N,ALIGNMENT,(uint_fast64_t)args->buf_mt_seed));
            }else{
                buf_multi.reset(new MemoryBufferMulti(alloc.get(),args->num_multi_buf,N,ALIGNMENT));
            }
            buf_multi->fill();
        }else{
            if(args->randomize_buffers){
                buf1.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT,(uint_fast64_t)args->buf_mt_seed));
            }else{
                buf1.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT));
            }
            buf1->fill();
        }
        if(args->randomize_buffers){
            buf2.reset(new MemoryBuffer(alloc.get(),0,ALIGNMENT,(uint_fast64_t)args->buf_mt_seed));
        }else{
            buf2.reset(new MemoryBuffer(alloc.get(),0,ALIGNMENT));
        }
        buf2->fill();
    }else{
        if(args->randomize_buffers){
            buf1.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT,(uint_fast64_t)args->buf_mt_seed));
            buf2.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT,(uint_fast64_t)args->buf_mt_seed));
        }else{
            buf1.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT));
            buf2.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT));
        }
        buf1->fill();
        buf2->fill();
    }
    if (args->do_alltoall){ //All-to-All
        N=args->len_msg*cl->size()+cl->kpingpong_minimal_buffer_overhead();
        if(args->randomize_buffers){
            buf_all2all.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT,(uint_fast64_t)args->buf_mt_seed));
        }else{
            buf_all2all.reset(new MemoryBuffer(alloc.get(),N,ALIGNMENT));
        }
        buf_all2all->fill();
    }
}

void Benchmark::run_Stresstest() {
    computeRankStepMappings();
    LinktestStats dummy;
    stats = &dummy;

    double timeLastIteration, timeWork=0.0, timeTotal=0.0; // Total time = work time + all-to-all time (+ overhead)
    for(std::size_t iteration = 1;; iteration++) {
        timeLastIteration = timeTotal;
        run_iteration(iteration, std::numeric_limits<std::size_t>::max(), &timeWork, &timeTotal);
        timeLastIteration = timeTotal - timeLastIteration;
        if (isLastIteration(timeTotal, timeLastIteration, iteration)) {
            break;
        } else {
            if(cl->rank() == 0) {
                std::printf("LinkTest iteration finished, stress test continuing: elapsed time=%ss, finished iterations=%zu<=%" PRIu64 "\n",
                    UnitPrefix::SI_prefix(timeTotal, 10).get(),
                    iteration,
                    args->min_iterations
                );
                std::fflush(stdout);
            }
        }
    }
}

void Benchmark::run_Measuringtest() {
    const std::size_t NUM_ITERATIONS = randomizeTasks() ? args->num_randomize_tasks : 1;
    newStatsVec.resize(NUM_ITERATIONS);

    double timeWork, timeTotal=0.0; // Total time = work time + all-to-all time (+ overhead) for each iteration
    for(std::size_t iteration = 1; iteration <= NUM_ITERATIONS; iteration++) {
        if (0 == cl->rank()) {
            std::printf("\n\nIteration:%4ld of %4ld\n", iteration, NUM_ITERATIONS);
            std::printf("Preparing\n");
        }
        timeWork=0;
        computeRankStepMappings();
        stats = &newStatsVec.at( iteration - 1 );
        stats->initialize(cl.get(), args);
        if (0 == cl->rank()) {
            std::printf("Starting\n");
            std::printf(HLINE);
        }
        if (0 == cl->rank()) stats->setStartTimeString();
        run_iteration(iteration, NUM_ITERATIONS-iteration, &timeWork, &timeTotal);
        if (0 == cl->rank()) stats->setEndTimeString();
        handle_slow_pairs(iteration);
    }
    write_to_file();
}

/* The benchmark consists of two nested loops. The outer loop is used
 * to ensure a minimal amount of iterations and a minimal runtime.
 */
int Benchmark::benchmark(){
    init();

    std::unique_ptr<StopwatchI> rootWatch = Stopwatchfactory::getRootWatch(cl->rank());
    rootWatch->start();
    prepareBuffers();

    rootWatch->start();
    cl->barrier();

    rootWatch->stop();
    printTimingIfRoot(cl->rank(), "[first sync]", rootWatch->getDuration());

    if(rank() == 0) print_cmdline_args(args);

    if( isStressTest() ) {
        run_Stresstest();
    } else {
        run_Measuringtest();
    }

    rootWatch->stop();
    printTimingIfRoot(cl->rank(), "[all]", rootWatch->getDuration());

    return SUCCESS;
}

}
